/*!
 *  \file   ITMtrace.cpp
 *  \brief  Trace for ITM on Cortex-M
 *  \date   Copyright ARM Limited 2009 All Rights Reserved.
 *
 */

/*
	Notes:
		When a parameter is changed, this plugin is notified immediatly through the CADI function 'SetParameterValues'
		'SetParameterValues' then calls 'SetParametersInternal' which calls 'SetTraceFile'
		'SetTraceFile' either opens stdout (if filename is blank) or opens 'ITM_Ports' number of files 

	What to add for Cirrus:
	(externally in Python) 1) Add ability to dynamically change file names
	1.5) Make file names have _#, not just # after them
	2) Added string old_trace_file_name which stores the previously used file name
	3) Upon closing ITMtrace OR changing the filename parameter mid-test:
		clear all empty files (using the old_trace_file_name string if mid-test) with the DeleteEmptyFile function
		IF only one file (port) has been used, strip the _# extention
*/

#include "ITMtrace.h"

// Initalize this parameter up here for easy readability

string old_trace_file_name;
int file_filled_tracker = 0;  //tracks the number of files that are non-empty
string file_thats_filled= ""; //tracks the name of the last file that is filled. 

ITMtrace::ITMtrace(const char *instance_name_,
                                       uint32_t number_parameters,
                                       eslapi::CADIParameterValue_t *parameter_values) :
    instance_name(instance_name_),
    itm_payload_index(-1),
    itm_hdr_index(-1),
    interfaceRegistry("ITMtrace")
{
    printf("ITMtrace: plugin instance name: %s\n",instance_name_);
	if (number_parameters > 0)
    {
        SetParametersInternal(number_parameters, parameter_values);
    }

    interfaceRegistry.Register<PluginInstance>(this);
    interfaceRegistry.Register<ParameterInterface>(this);

}

ITMtrace::~ITMtrace()
{ 

    // close itm port trace files currently open. 
    (void) CloseTraceFile();

	// reset vars
	file_filled_tracker = 0;
	file_thats_filled = "";

	// old_trace_file_name is global and already defined....split it into body and extention
	size_t pos = old_trace_file_name.find_last_of(".");  // position of start of file extention
	string f_body;
	string f_extention;

	// if no '.' is found, add a .txt extention:
	if (pos == string::npos) {
		f_body = old_trace_file_name;
		f_extention = ".txt";
	}
	// otherwise, split the file name into the body and extention (hello.txt gets split into 'hello' and '.txt'
	else {
		f_body = old_trace_file_name.substr(0, pos);
		f_extention = old_trace_file_name.substr(pos);
	}

	//printf("ITM~: %d %s %s\n", pos, f_body, f_extention);


	for (int i = 0; i < ITM_PORTS; i++)
	{
		string f_name = f_body + "_ITM" + std::to_string(i) + f_extention;
		(void)DeleteEmptyFile(f_name.c_str());
	}

	// if file_filled_tracker is 1, remove the _ITM# after it. 
	if (file_filled_tracker == 1)
	{
		int x = file_thats_filled.find('.'); // Gets position of '.' in file name string
		int y = file_thats_filled.find_last_of('_');

		std::string new_file_name = file_thats_filled.substr(0, y) + file_thats_filled.substr(x);  // cut out _ITM# from name
		std::rename(file_thats_filled.c_str(), new_file_name.c_str());
	}

	// clean up
	file_filled_tracker = 0;
	file_thats_filled = "";


	
    printf("ITMtrace: stopping plugin\n");

}

eslapi::CAInterface *
ITMtrace::ObtainInterface(eslapi::if_name_t    ifName,
                                    eslapi::if_rev_t     minRev,
                                    eslapi::if_rev_t *   actualRev)
{
    return interfaceRegistry.ObtainInterface(ifName, minRev, actualRev);
}

eslapi::CADIReturn_t
ITMtrace::Error(const char *message) const
{
    fprintf(stderr, "Error: Trace plugin %s: %s\n", instance_name.c_str(), message);
    return eslapi::CADI_STATUS_GeneralError;
}

eslapi::CADIReturn_t
ITMtrace::RegisterSimulation(eslapi::CAInterface *ca_interface)
{
    if (!ca_interface)
        return eslapi::CADI_STATUS_IllegalArgument;

    // Get the System Trace Interface:
    MTI::SystemTraceInterface *sti = ca_interface->ObtainPointer<MTI::SystemTraceInterface>();

    if (!sti)
        return eslapi::CADI_STATUS_IllegalArgument;

    // Check if there is at least one component with trace:
    MTI::SystemTraceInterface::TraceComponentIndex num_trace_components = sti->GetNumOfTraceComponents();
    if (num_trace_components == 0)
        return Error("No components provide trace.");


    // For each component get the Trace Component Interface and see if it has an INST source
    for(MTI::SystemTraceInterface::TraceComponentIndex tci=0; tci < num_trace_components; ++tci)
    {
        eslapi::CAInterface *caif = sti->GetComponentTrace(tci);
        MTI::ComponentTraceInterface *cti = caif->ObtainPointer<MTI::ComponentTraceInterface>();
        if(cti)
        {
            // Get the trace source named "ITM":
            MTI::TraceSource *source = cti->GetTraceSource("ITM");
            if (source)
            {
                printf("ITMtrace: attached to component: %s\n", sti->GetComponentTracePath(tci));

                // Now find the field payload, and create a field mask:
                const MTI::EventFieldType *field = source->GetField("PACKET_PAYLOAD");
                if (!field)
                    return Error("No field named \"PACKET_PAYLOAD\" found in \"ITM\" trace source.");
                MTI::FieldMask mask = 1 << field->GetIndex();

                field = source->GetField("PACKET_HEADER");
                if (!field)
                    return Error("No field named \"PACKET_HEADER\" found in \"ITM\" trace source.");

                mask |= 1 << field->GetIndex();

                // Register an event class:
                event_class = source->CreateEventClass(mask);
                if (!event_class)
                    return Error("Problem when creating EventClass.");

                // Now register the callback:
                MTI::Status status = event_class->RegisterCallback(TraceITM_Thunk, this);
                if (status != MTI::MTI_OK)
                    return Error("EventClass::RegisterCallback() returned error.");

                itm_payload_index = event_class->GetValueIndex("PACKET_PAYLOAD");
                if (itm_payload_index == -1)
                    return Error("EventClass::GetValueIndex(\"PACKET_PAYLOAD\") returned error.");
                itm_hdr_index = event_class->GetValueIndex("PACKET_HEADER");
                if (itm_hdr_index == -1)
                    return Error("EventClass::GetValueIndex(\"PACKET_HEADER\") returned error.");
            }
        }
    }
    return eslapi::CADI_STATUS_OK;
}

// This is called before the plugin .dll/.so is unloaded and should allow the plugin to do it's cleanup.
void
ITMtrace::Release()
{
    delete this;
}

const unsigned int num_parameters = 1;

typedef enum
{
    PARAM_ID_TRACE_FILE
} MyParameterIDs;

static CADIParameterInfo_t parameters[num_parameters] =
{
    CADIParameterInfo_t(PARAM_ID_TRACE_FILE,
                        "trace-file",
                        CADI_PARAM_STRING,
                        "The trace file to write into. If empty will print to stdout.",
                        true, // runtime
                        0, 0, 0,
                        ""), // default: STDOUT
};


const char *
ITMtrace::GetName() const
{
    return instance_name.c_str();
}

/////////////////////////////////////////////////////////////////////////////

void ITMtrace::TraceITM(const MTI::EventClass *event_class,
                                 const struct MTI::EventRecord *record)
{
    uint32_t payload = record->Get<uint32_t>(event_class, itm_payload_index);
    uint32_t header = record->Get<uint32_t>(event_class, itm_hdr_index);
    uint32_t size = (header & 3);
    uint32_t port = ((header & 0xf8) >> 3);

    if (size == 1) // 1 byte payload
       fprintf(trace_file[port], "%c", payload);
    else  // 2 or 4 byte payload
       fprintf(trace_file[port], "%c",payload);

    fflush(trace_file[port]);
}

// The callback interface is a C interface, so we need a short thunk
// function to call into a C++ method. This is done via a static C++
// method, which behaves like a C function:
void
ITMtrace::TraceITM_Thunk(void * user_data,
                                  const MTI::EventClass *event_class,
                                  const struct MTI::EventRecord *record)
{
    ITMtrace *simple_trace = reinterpret_cast<ITMtrace *>(user_data);
    simple_trace->TraceITM(event_class, record);
}

/////////////////////////////////////////////////////////////////////////////

void
ITMtrace::DeleteEmptyFile(const char *filename) 
{
    struct stat st;

	if(stat(filename, &st) != 0) {
        return;
    }

    if (st.st_size == 0)
        std::remove(filename); // delete empty file
	else {
		file_filled_tracker += 1;
		file_thats_filled = filename; 
	}
}
void
ITMtrace::CloseTraceFile()
{
    if (trace_file[0] == stdout)
    {
        fflush(stdout);
        trace_file[0] = (FILE *) NULL;
    }
    else
    {
        for (int i = 0; i < ITM_PORTS; i++)
        {
            fclose(trace_file[i]);
            trace_file[i] = (FILE *) NULL;
        }
    }
}

void 
ITMtrace::SetTraceFile(const char *new_filename)
{
    trace_file_name = new_filename;

	size_t pos = trace_file_name.find_last_of(".");  // position of start of file extention
	string f_body;
	string f_extention;
	string old_f_body;
	string old_f_extention;

    if (trace_file_name.empty())
    {
		printf("ITMtrace: set ITM output to: stdout\n");

		// set old_trace_file_name to blank
		old_trace_file_name = "";
        for (int i = 0; i < ITM_PORTS; i++)
            trace_file[i] = stdout;
    }
    else
    {
		printf("ITMtrace: set ITM output to: %s\n", trace_file_name.c_str());

		// if no '.' is found, add a .txt extention:
		if (pos == string::npos) {
			f_body = trace_file_name;
			f_extention = ".txt";
		}
		// otherwise, split the file name into the body and extention (hello.txt gets split into 'hello' and '.txt'
		else {
			f_body = trace_file_name.substr(0, pos);
			f_extention = trace_file_name.substr(pos);
		}

		//printf("ITM: %d %s %s\n", pos, f_body, f_extention);

		// If there were trace files before, close them before opening new ones. 
		if (!old_trace_file_name.empty()) {
			// get old trace file body and extention
			// if no '.' is found, add a .txt extention:
			if (pos == string::npos) {
				old_f_body = old_trace_file_name;
				old_f_extention = ".txt";
			}
			// otherwise, split the file name into the body and extention (hello.txt gets split into 'hello' and '.txt'
			else {
				old_f_body = old_trace_file_name.substr(0, pos);
				old_f_extention = old_trace_file_name.substr(pos);
			}

			// close itm port trace files currently open. 
			(void)CloseTraceFile();

			////////// delete them
			file_filled_tracker = 0;
			file_thats_filled = "";

			for (int i = 0; i < ITM_PORTS; i++)
			{
				string f_name = old_f_body + "_ITM" + std::to_string(i) + old_f_extention;
				(void)DeleteEmptyFile(f_name.c_str());
			}

			// if file_filled_tracker is 1, remove the _ITM# after it to make it more clean
			if (file_filled_tracker == 1)
			{
				int x = file_thats_filled.find('.'); // Gets position of '.' in file name string
				int y = file_thats_filled.find_last_of('_');

				std::string new_file_name = file_thats_filled.substr(0, y) + file_thats_filled.substr(x);  // cut out _ITM# from name
				std::rename(file_thats_filled.c_str(), new_file_name.c_str());
			}

			// clean up
			file_filled_tracker = 0;
			file_thats_filled = "";
		}
		//set old_trace_file_name to the current trace_file_name
		old_trace_file_name = trace_file_name;

		//open ports
        for (int i = 0; i < ITM_PORTS; i++)
        {
            //string file_name = trace_file_name + "_ITM" + std::to_string(i);
			string file_name = f_body + "_ITM" + std::to_string(i) + f_extention;
			trace_file[i] = fopen(file_name.c_str(), "wb");
        }
    }
}

CADIReturn_t
ITMtrace::GetParameterInfos(uint32_t startIndex,
                                     uint32_t desiredNumOfParams,
                                     uint32_t *actualNumOfParams,
                                     CADIParameterInfo_t *params)
{
    if (!actualNumOfParams || !params)
        return CADI_STATUS_IllegalArgument;

    uint32_t i = 0;
    for(; i < desiredNumOfParams; ++i)
    {
        if (startIndex + i >= num_parameters)
            break;
        params[i] = parameters[startIndex + i];
    }
    *actualNumOfParams = i;

    return CADI_STATUS_OK;
}

CADIReturn_t
ITMtrace::GetParameterInfo(const char *parameterName,
                                  CADIParameterInfo_t *param)
{
    if (!param)
        return CADI_STATUS_IllegalArgument;

    for(uint32_t i=0; i < num_parameters; ++i)
    {
        if (strcmp(parameters[i].name, parameterName) == 0)
        {
            *param = parameters[i];
            return CADI_STATUS_OK;
        }
    }

    return CADI_STATUS_ArgNotSupported; // not found
}

CADIReturn_t
ITMtrace::GetParameterValues(uint32_t parameterCount,
                                 uint32_t *actualNumOfParamsRead,
                                 CADIParameterValue_t *paramValuesOut)
{
    if (!paramValuesOut)
        return CADI_STATUS_IllegalArgument;

    for(uint32_t i = 0; i < parameterCount; ++i)
    {
        CADIParameterValue_t &param_value = paramValuesOut[i];
        switch(param_value.parameterID)
        {
        case PARAM_ID_TRACE_FILE:
            AssignString(param_value.stringValue, trace_file_name.c_str(), sizeof(param_value.stringValue));
            break;

        default: // unknown ID
            if (actualNumOfParamsRead) 
                *actualNumOfParamsRead = i;
            return CADI_STATUS_IllegalArgument;
        }
    }

    if (actualNumOfParamsRead)
        *actualNumOfParamsRead = parameterCount;

    return CADI_STATUS_OK;
}



CADIReturn_t
ITMtrace::SetParameterValues(uint32_t parameterCount,
                                 CADIParameterValue_t *parameter_values,
                                 CADIFactoryErrorMessage_t * /*error*/)
{
    if (!parameter_values)
        return CADI_STATUS_IllegalArgument;
    return SetParametersInternal(parameterCount, parameter_values);
}


CADIReturn_t
ITMtrace::SetParametersInternal(uint32_t parameterCount, CADIParameterValue_t *parameter_values)
{

    for(uint32_t i = 0; i < parameterCount; ++i)
    {
        CADIParameterValue_t &parameter = parameter_values[i];

        switch(parameter.parameterID)
        {
            case PARAM_ID_TRACE_FILE:
				// close existing itm port trace files if you are opening new ones
				//(void)CloseTraceFile();

                SetTraceFile(parameter.stringValue);
            break;

        }
    }

    return CADI_STATUS_OK;
}



eslapi::CAInterface *
ThePluginFactory::ObtainInterface(eslapi::if_name_t    ifName,
                                  eslapi::if_rev_t     minRev,
                                  eslapi::if_rev_t *   actualRev)
{
    return interfaceRegistry.ObtainInterface(ifName, minRev, actualRev);
}

uint32_t
ThePluginFactory::GetNumberOfParameters()
{
    return num_parameters;
}


eslapi::CADIReturn_t
ThePluginFactory::GetParameterInfos(eslapi::CADIParameterInfo_t* parameter_info_list)
{
    if (!parameter_info_list)
        return CADI_STATUS_IllegalArgument;

    std::copy( parameters, &parameters[num_parameters], parameter_info_list);
    return CADI_STATUS_OK;
}


eslapi::CAInterface *
ThePluginFactory::Instantiate(const char *instance_name,
                              uint32_t    number_of_parameters,
                              eslapi::CADIParameterValue_t *param_values)
{
    return static_cast<MTI::PluginInstance *>(new ITMtrace(instance_name, number_of_parameters, param_values));
}

void
ThePluginFactory::Release()
{
    // nothing to do since factory is a static instance
}

static ThePluginFactory factory_instance;

extern "C"
{
    eslapi::CAInterface *
    GetCAInterface()
    {
        return &factory_instance;
    }
}
// End of file ITMtrace.cpp
